use serde::{Deserialize, Serialize};
use std::sync::Arc;

use axum::{
    extract::{
        ws::{Message, WebSocket, WebSocketUpgrade},
        State,
    },
    response::IntoResponse,
};

use dist_text::crdts::list::Op;
use futures::{sink::SinkExt, stream::StreamExt};

use crate::AppState;

#[derive(Serialize, Deserialize)]
struct NewState {
    client: String,
    doc: String,
    ops: Vec<Op<u16, String>>,
}

#[derive(Clone, Serialize, Deserialize)]
pub(crate) struct Ops {
    pub(crate) client: String,
    ops: Vec<Op<u16, String>>,
}

pub(crate) async fn route(
    ws: WebSocketUpgrade,
    State(state): State<Arc<AppState>>,
) -> impl IntoResponse {
    ws.on_upgrade(move |socket| handle_socket(socket, state))
}

async fn handle_socket(stream: WebSocket, state: Arc<AppState>) {
    let (mut sender, mut receiver) = stream.split();

    {
        let doc = state.doc.read().await;
        let doc: String = doc.to_string();
        let new_state = NewState {
            client: String::from("0000"),
            doc: doc.to_string(),
            ops: Vec::new(),
        };
        let payload = serde_json::to_string(&new_state).unwrap();
        let _ = sender.send(Message::Text(payload)).await;
    }

    let mut rx = state.tx.subscribe();

    let mut send_task = tokio::spawn(async move {
        while let Ok(msg) = rx.recv().await {
            if sender.send(Message::Text(msg)).await.is_err() {
                break;
            }
        }
    });

    let tx = state.tx.clone();
    let state = state.clone();
    let mut recv_task = tokio::spawn(async move {
        while let Some(Ok(Message::Text(text))) = receiver.next().await {
            if let Ok(new_ops) = serde_json::from_str::<Ops>(&text) {
                tracing::debug!("update from {}", new_ops.client.clone());
                let doc = state.doc.read().await;
                let mut doc = doc.to_owned();

                doc.apply_ops(new_ops.ops.to_owned());

                let new_state = NewState {
                    client: new_ops.client,
                    doc: doc.to_string(),
                    ops: new_ops.ops,
                };
                let payload = serde_json::to_string(&new_state).unwrap();
                let _ = tx.send(payload);
            }
        }
    });

    tokio::select! {
        _ = (&mut send_task) => recv_task.abort(),
        _ = (&mut recv_task) => send_task.abort(),
    };
}
